__doc__ = """
Module for extracting each spike waveform and visualize.
"""
__all__ = ["ExtractWaveforms"]

from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import inspect
import os
import pathlib
from dataclasses import dataclass

import matplotlib
import matplotlib.pyplot as plt
import neo
import numpy as np
import quantities as pq
from scipy.signal import lfilter, savgol_filter
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.wrapper import wrap_generator_to_generator
from miv.typing import SignalType, SpikestampsType


@dataclass
class ExtractWaveforms(OperatorMixin):
    """
    Extract spike waveforms as signal cutouts around each spike index as a spikes x samples numpy array

    Parameters
    ----------
    channels : Optional[List[int]]
        Interested channels. If None, extract from all channels. (default: None)
    pre : pq.Quantity | float
        The duration of the cutout before the spike in seconds. (default: 0.001 s)
    post : pq.Quantity | float
        The duration of the cutout after the spike in seconds. (default: 0.002 s)
    plot_n_spikes : Optional[int]
        The number of cutouts to plot. None to plot all. (default: 100)
    """

    channels: Optional[List[int]] = None
    pre: pq.Quantity = 0.001 * pq.s
    post: pq.Quantity = 0.002 * pq.s
    plot_n_spikes: int = 100
    tag: str = "extract waveform"

    def __post_init__(self):
        super().__init__()

    def __call__(
        self, signal: Generator[Signal, None, None], spikestamps: Spikestamps
    ) -> Dict[int, Signal]:
        """__call__

        Parameters
        ----------
        signal : Generator[Signal, None, None]
            The signal to extract the event feature
        spikestamps : Spikestamps
            The sample index of all spikes as a 1-dim numpy array

        Returns
        -------
        Stack of spike cutout: List[np.ndarray]
            Return stacks of spike cutout; shape(n_spikes, width).

        """

        if isinstance(self.pre, pq.Quantity):
            pre = self.pre.rescale(pq.s).magnitude
        else:
            pre = self.pre
        if isinstance(self.post, pq.Quantity):
            post = self.post.rescale(pq.s).magnitude
        else:
            post = self.post
        if not inspect.isgenerator(signal):
            signal = [signal]

        cutouts = {}
        sampling_rate = None
        for sig in signal:  # TODO: Use generator-accepted caller instead
            sampling_rate = sig.rate
            num_channels = sig.number_of_channels
            channels = range(num_channels) if self.channels is None else self.channels
            pre_idx = int(pre * sampling_rate)
            post_idx = int(post * sampling_rate)
            assert (
                pre_idx + post_idx > 0
            ), "Set larger pre/post duration. pre+post duration must be more than 1/sampling_rate."
            spikestamps_view = spikestamps.get_view(
                sig.get_start_time(), sig.get_end_time()
            )
            for ch in channels:
                # Padding signal
                spikestamp = spikestamps_view[ch]
                if len(spikestamp) == 0:
                    continue
                padded_signal = np.pad(
                    sig.data[:, ch], ((pre_idx, post_idx),), constant_values=0
                )
                cutout = np.empty(
                    [pre_idx + post_idx, len(spikestamp)], dtype=np.float_
                )
                for idx, time in enumerate(spikestamp):
                    index = int(round((time - sig.get_start_time()) * sampling_rate))
                    if (
                        index >= sig.data.shape[0]
                        or index + post_idx + pre_idx >= sig.data.shape[0]
                    ):
                        raise IndexError(
                            f"The width of the spike {post_idx+pre_idx} exceeded the signal {sig.data.shape[0]}. "
                            "Either timestamp exceeded the maximum time recorded, or "
                            "post duration is too large."
                        )
                    cutout[:, idx] = padded_signal[index : (index + post_idx + pre_idx)]
                if ch not in cutouts:
                    cutouts[ch] = Signal(
                        data=cutout,
                        timestamps=np.arange(pre_idx + post_idx).astype(np.float_)
                        / sampling_rate
                        - pre,
                        rate=sampling_rate,
                    )
                else:
                    cutouts[ch].append(cutout)

        return cutouts

    def plot_waveforms(
        self,
        cutouts: Dict[int, Signal],
        show: bool = False,
        save_path: Optional[pathlib.Path] = None,
        plot_kwargs: Dict[Any, Any] = None,
    ) -> plt.Figure:
        """
        Plot an overlay of spike cutouts

        Parameters
        ----------
        cutouts : np.ndarray
            A spikes x samples array of cutouts
        plot_kwargs : Dict[Any, Any]
            Addtional keyword-arguments for matplotlib.pyplot.plot.
        """
        output_axes = []
        for ch, signal in cutouts.items():
            cutout = signal.data
            num_cutout = signal.number_of_channels
            if self.plot_n_spikes is None:
                self.plot_n_spikes = num_cutout
            plot_n_spikes = min(self.plot_n_spikes, num_cutout)
            if num_cutout == 0:
                continue

            if not plot_kwargs:
                plot_kwargs = {"alpha": 0.3, "linewidth": 1, "color": "b"}

            time = signal.timestamps * pq.s
            time = time.rescale(pq.ms).magnitude

            fig, ax = plt.subplots(figsize=(8, 5))
            for i in range(plot_n_spikes):
                ax.plot(
                    time,
                    cutout[:, i],
                    **plot_kwargs,
                )
            ax.set_xlabel("Time (ms)")
            ax.set_ylabel("Voltage (mV)")
            ax.set_title(f"Spike Cutouts (channel {ch})")
            output_axes.append(ax)
            if save_path:
                fig.savefig(os.path.join(save_path, f"spike_cutouts_ch{ch:03}.png"))
        if show:
            plt.show()

        return output_axes



__all__ = ["PythonDataType", "NumpyDType"]

from typing import Protocol, Union

import numpy as np

from miv.core.operator.chainable import BaseChainingMixin


class RawValuesProtocol(Protocol):
    @staticmethod
    def is_valid(value) -> bool:
        ...


class ValuesMixin(BaseChainingMixin):
    """
    This mixin is used to convert pure/numpy data type to be a valid input/output of a node.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output = value

    @property
    def output(self):
        return self._output

    def run(self, **kwargs):
        pass


class PythonDataType(ValuesMixin):
    """
    Python data types are: int, float, str, bool, list, tuple, dict
    This type also allows None.
    """

    @staticmethod
    def is_valid(value):
        return value is None or isinstance(
            value, (int, float, str, bool, list, tuple, dict)
        )


class NumpyDType(ValuesMixin):
    """
    Numpy data types.
    """

    @staticmethod
    def is_valid(value):
        return isinstance(value, np.ndarray)



__all__ = ["ParallelGeneratorFetch"]

import os
import sys
from multiprocessing import Process, Queue
from queue import Empty


class ExceptionItem:
    def __init__(self, exception):
        self.exception = exception


class ParallelGeneratorException(Exception):
    pass


class ParallelGeneratorFetch:
    def __init__(self, orig_gen, max_lookahead=None, get_timeout=10):
        """
        Creates a parallel generator from a normal one.
        The elements will be prefetched up to max_lookahead
        ahead of the consumer. If max_lookahead is None,
        everything will be fetched.

        The get_timeout parameter is the number of seconds
        after which we check that the subprocess is still
        alive, when waiting for an element to be generated.

        Any exception raised in the generator will
        be forwarded to this parallel generator.
        """
        if max_lookahead:
            self.queue = Queue(max_lookahead)
        else:
            self.queue = Queue()

        def wrapped():
            try:
                for item in orig_gen:
                    self.queue.put(item)
                raise StopIteration()
            except Exception as e:
                self.queue.put(ExceptionItem(e))

        self.get_timeout = get_timeout
        self.ppid = None  # pid of the parent process
        self.process = Process(target=wrapped)
        self.process_started = False

    def finish_if_possible(self):
        """
        We can only terminate the child process from the parent process
        """
        if self.ppid == os.getpid() and self.process:  # and self.process.is_alive():
            self.process.terminate()
            self.process = None
            self.queue = None
            self.ppid = None

    def __enter__(self):
        """
        Starts the process
        """
        self.ppid = os.getpid()
        self.process.start()
        self.process_started = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Kills the process
        """
        assert self.process_started and self.ppid == None or self.ppid == os.getpid()
        self.finish_if_possible()

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self

    def __del__(self):
        self.finish_if_possible()

    def next(self):
        if not self.process_started:
            raise ParallelGeneratorException(
                """The generator has not been started.
                   Please use "with ParallelGenerator(..) as g:"
                """
            )
        try:
            item_received = False
            while not item_received:
                try:
                    item = self.queue.get(timeout=self.get_timeout)
                    item_received = True
                except Empty:
                    # check that the process is still alive
                    if not self.process.is_alive():
                        raise ParallelGeneratorException(
                            "The generator died unexpectedly."
                        )

            if type(item) == ExceptionItem:
                raise item.exception
            return item

        except Exception:
            self.finish_if_possible()
            raise



__all__ = ["Extendable"]

from typing import List, Protocol, TypeVar


class Extendable(Protocol):
    def extend(self, other) -> None:
        ...


ChannelWiseSelf = TypeVar("ChannelWiseSelf", bound="ChannelWise")


class ChannelWise(Protocol):
    @property
    def number_of_channels(self) -> int:
        ...

    def append(self, other) -> None:
        ...

    def insert(self, index, other) -> None:
        ...

    def __setitem__(self, index, other) -> None:
        ...

    def select(self, indices: List[int]) -> ChannelWiseSelf:
        """
        Select channels by indices.
        """
        ...



from __future__ import annotations

__doc__ = """
"""
__all__ = [
    "_CacherProtocol",
    "_Jsonable",
    "_Cachable",
    "DataclassCacher",
    "FunctionalCacher",
]

from typing import TYPE_CHECKING, Any, Generator, Literal, Protocol, Union

import collections
import dataclasses
import functools
import glob
import itertools
import json
import os
import pathlib
import pickle as pkl

import numpy as np

if TYPE_CHECKING:
    from miv.core.datatype import DataTypes

CACHE_POLICY = Literal["AUTO", "ON", "OFF", "MUST"]


class _CacherProtocol(Protocol):
    @property
    def cache_dir(self) -> str | pathlib.Path:
        ...

    @property
    def cache_tag(self) -> str:
        ...

    @property
    def config_filename(self) -> str | pathlib.Path:
        ...

    def cache_filename(self) -> str | pathlib.Path:
        ...

    def load_cached(self) -> Generator[Any, None, None]:
        """Load the cached values."""
        ...

    def save_cache(self, values: Any, idx: int) -> bool:
        ...

    def save_config(self) -> None:
        ...

    def check_cached(self) -> bool:
        """Check if the current configuration is the same as the cached one."""
        ...


class _Jsonable(Protocol):
    def to_json(self) -> dict[str, Any]:
        ...


class _Cachable(Protocol):
    @property
    def analysis_path(self) -> str | pathlib.Path:
        ...

    @property
    def cacher(self) -> _CacherProtocol:
        ...

    def set_caching_policy(self, policy: CACHE_POLICY) -> None:
        ...

    def run(self, cache_dir: str | pathlib.Path) -> None:
        ...


class SkipCache:  # TODO
    """
    Always run without saving.
    """

    def __init__(self, parent, cache_dir: str | pathlib.Path):
        super().__init__()

    @property
    def config_filename(self) -> str:
        raise NotImplementedError(
            "If you are using SkipCache, you should not be calling this method."
        )

    def cache_filename(self, idx) -> str:
        raise NotImplementedError(
            "If you are using SkipCache, you should not be calling this method."
        )

    def check_cached(self) -> bool:
        return False

    def save_config(self):
        raise NotImplementedError(
            "If you are using SkipCache, you should not be calling this method."
        )

    def load_cached(self):
        raise NotImplementedError(
            "If you are using SkipCache, you should not be calling this method."
        )

    def save_cache(self, values, idx):
        raise NotImplementedError(
            "If you are using SkipCache, you should not be calling this method."
        )


def when_policy_is(*allowed_policy):
    def decorator(func):
        # @functools.wraps(func) # TODO: fix this
        def wrapper(self, *args, **kwargs):
            if self.policy in allowed_policy:
                return func(self, *args, **kwargs)
            else:
                return False

        return wrapper

    return decorator


def when_initialized(func):  # TODO: refactor
    # @functools.wraps(func) # TODO: fix this
    def wrapper(self, *args, **kwargs):
        if self.cache_dir is None:
            return False
        else:
            return func(self, *args, **kwargs)

    return wrapper


class BaseCacher:
    def __init__(self, parent):
        super().__init__()
        self.policy: CACHE_POLICY = "AUTO"  # TODO: make this a property
        self.parent = parent
        self.cache_dir = None  # TODO: Public. Make proper setter
        self.cache_tag = "data"

    @property
    def config_filename(self) -> str:
        return os.path.join(self.cache_dir, "config.json")

    def cache_filename(self, idx) -> str:
        index = idx if isinstance(idx, str) else f"{idx:04}"
        return os.path.join(self.cache_dir, f"cache_{self.cache_tag}_{index}.pkl")

    @when_policy_is("ON", "AUTO", "MUST")
    @when_initialized
    def save_cache(self, values, idx=0) -> bool:
        os.makedirs(self.cache_dir, exist_ok=True)
        with open(self.cache_filename(idx), "wb") as f:
            pkl.dump(values, f)
        return True


class DataclassCacher(BaseCacher):
    @when_policy_is("ON", "AUTO", "MUST")
    @when_initialized
    def check_cached(self) -> bool:
        if self.policy == "MUST":
            return True
        current_config = self._compile_configuration_as_dict()
        cached_config = self._load_configuration_from_cache()
        if cached_config is None:
            flag = False
        else:
            flag = current_config == cached_config  # TODO: fix this
        return flag

    def _load_configuration_from_cache(self) -> dict:
        if os.path.exists(self.config_filename):
            with open(self.config_filename) as f:
                return json.load(f)
        return None

    def _compile_configuration_as_dict(self) -> dict:
        config = dataclasses.asdict(self.parent, dict_factory=collections.OrderedDict)
        for key in config.keys():
            if isinstance(config[key], np.ndarray):
                config[key] = config[key].tostring()
            elif hasattr(config[key], "to_json"):
                config[key] = config[key].to_json()
        return config

    @when_policy_is("ON", "AUTO", "MUST")
    @when_initialized
    def save_config(self):
        config = self._compile_configuration_as_dict()
        os.makedirs(self.cache_dir, exist_ok=True)
        try:
            with open(self.config_filename, "w") as f:
                json.dump(config, f, indent=4)
        except (TypeError, OverflowError):
            raise TypeError(
                "Some property of caching objects are not JSON serializable."
            )
        return True

    @when_initialized
    def load_cached(self) -> Generator[DataTypes, None, None]:
        paths = glob.glob(self.cache_filename("*"))
        for path in paths:
            with open(path, "rb") as f:
                yield pkl.load(f)


class FunctionalCacher(BaseCacher):
    @when_policy_is("ON", "AUTO", "MUST")
    @when_initialized
    def check_cached(self) -> bool:
        if self.policy == "MUST":  # TODO: fix this, remove redundancy
            return True
        flag = os.path.exists(self.cache_filename(0))
        return flag

    @when_policy_is("ON", "AUTO", "MUST")
    @when_initialized
    def save_config(self):
        pass

    @when_initialized
    def load_cached(self) -> Generator[DataTypes, None, None]:
        path = glob.glob(self.cache_filename(0))[0]
        with open(path, "rb") as f:
            yield pkl.load(f)



__all__ = ["PSTH", "peri_stimulus_time", "PSTHOverlay"]

from typing import List

import itertools
import os
import sys
from dataclasses import dataclass
from glob import glob

import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator import OperatorMixin
from miv.core.wrapper import wrap_cacher
from miv.mea import mea_map


def peri_stimulus_time(spike_list):
    """
    Compute the peri-stimulus time of the given spike train.

    Examples
    --------

    How to draw Peri-stimulus time histogram (ISIH):

        >>> from miv.statistics import peri_stimulus_time
        >>> import matplotlib.pyplot as plt
        >>> pst = peri_stimulus_time(spikestamp)
        >>> plt.hist(pst)

    If one wants to get the bin-count, you can use `numpy.digitize` and `numpy.bincount`:

        >>> import numpy as np
        >>> max_time = spikestamps.max()
        >>> num_bins = 20
        >>> time_interval = np.linspace(0, max_time, num_bins)
        >>> digitized = np.digitize(pst, time_interval)
        >>> bin_counts = np.bincount(digitized)

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        interval: numpy.ndarray

    """

    peri_stimulus_times = np.sum(np.array(spike_list), 0)
    return peri_stimulus_times


@dataclass
class PSTH(OperatorMixin):
    # Histogram Configuration
    # Default: 400ms domain, 4ms binsize
    mea: str = None
    binsize: float = 0.004  # seconds
    interval: float = 0.4  # seconds
    tag: str = "peri-stimulus time histogram"

    stimulus_length = 0.010

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    # @wrap_cacher("psth")
    def __call__(self, events: Spikestamps, spikestamps: Spikestamps):
        # TODO: Change events datatype to be Event, not Spikestamps
        n_time = int(np.ceil(self.interval / self.binsize))
        time_axis = np.linspace(0, self.interval, n_time)
        psth = np.zeros((spikestamps.number_of_channels, n_time), dtype=np.float_)
        for t_start in events[0]:
            t_end = t_start + n_time * self.binsize
            bst = spikestamps.binning(
                self.binsize,
                t_start=t_start + self.stimulus_length,
                t_end=t_end + self.stimulus_length,
                return_count=False,
            )
            for channel in range(spikestamps.number_of_channels):
                psth[channel] += bst[channel][:n_time]
        psth /= len(events[0])
        psth /= self.binsize
        return Signal(data=psth.T, timestamps=time_axis, rate=1.0 / self.binsize)

    def plot_psth_in_grid_map(self, psth, show=False, save_path=None):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for channel in range(psth.number_of_channels):
            p = psth[channel]
            time = psth.timestamps
            if channel not in mea_map:
                continue
            w = np.where(mea_map == channel)
            r = w[0][0]
            c = w[1][0]
            axes[r][c].plot(time, p)
            axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("mean (channels) spike rate per bin")

        # Legend
        for i, j in itertools.product(range(nrow), range(ncol)):
            axes[i, j].legend()

        plt.suptitle("PSTH: stimulating electrode")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth.png"))

        return axes


@dataclass
class PSTHOverlay(OperatorMixin):
    # TODO: Experimental
    # Histogram Configuration
    # Default: 400ms domain, 4ms binsize
    mea: str = None
    tag: str = "peri-stimulus time histogram overlay"

    stimulus_length = 0.010

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    def __call__(self, *psths):
        return psths

    def plot_psth_in_grid_map(
        self,
        psths,
        show=False,
        save_path=None,
    ):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for idx, psth in enumerate(psths):
            for channel in range(psth.number_of_channels):
                p = psth[channel]
                time = psth.timestamps
                if channel not in mea_map:
                    continue
                w = np.where(mea_map == channel)
                r = w[0][0]
                c = w[1][0]
                axes[r][c].plot(time, p, label=f"PSTH {idx}")
                axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("mean spike/bin")

        # Legend
        for i, j in itertools.product(range(nrow), range(ncol)):
            axes[i, j].legend()

        plt.suptitle("PSTH: stimulating electrode")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth.png"))

        return axes

    def plot_psth_area_trend(
        self,
        psths,
        show=False,
        save_path=None,
    ):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for channel in range(psths[0].number_of_channels):
            if channel not in mea_map:
                continue
            w = np.where(mea_map == channel)
            r = w[0][0]
            c = w[1][0]

            hist = []
            for idx, psth in enumerate(psths):
                p = psth[channel]
                time = psth.timestamps
                hist.append(np.trapz(p, time))
            axes[r][c].plot(hist)
            axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("data points")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("Area under PSTH")

        plt.suptitle("PSTH Trend")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth_trend.png"))

        return axes



__all__ = ["UnstructuredMEA"]

from typing import Tuple

import matplotlib
import numpy as np
from scipy.interpolate import griddata

from miv.mea.base import MEAMixin


class UnstructuredMEA(MEAMixin):
    """
    A class representing a grid-based multi-electrode array (MEA).

    Example
    -------
    For example, you could create an instance of the GridMEA class like this:
    literal blocks::

        grid = np.arrange(9).reshape(3, 3)
        mea = GridMEA(grid)
    """

    def __init__(
        self,
        indices: np.ndarray,
        coordinates: np.ndarray,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert (
            indices.shape[0] == coordinates.shape[0]
        ), f"The number of indices ({indices.shape[0]}) and coordinates ({coordinates.shape[0]}) must be the same."
        self.indices = indices
        self.coordinates = coordinates

    def to_json(self) -> dict:
        """Return a JSON-serializable dictionary"""
        return {
            "indices": self.indices.tolist(),
            "coordinates": self.coordinates.tolist(),
        }

    def map_data(
        self, vector: np.ndarray, missing_value: float = 0.0
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        values = np.full(self.indices.shape, missing_value)
        for idx, value in enumerate(vector):
            if idx not in self.indices:
                continue
            values[self.indices == idx] = value

        xmin = np.min(self.coordinates[:, 0])
        xmax = np.max(self.coordinates[:, 0])
        ymin = np.min(self.coordinates[:, 1])
        ymax = np.max(self.coordinates[:, 1])
        delx = xmax - xmin
        dely = ymax - ymin
        x = np.linspace(xmin - 0.1 * delx, xmax + 0.1 * delx, 100)
        y = np.linspace(ymin - 0.1 * dely, ymax + 0.1 * dely, 100)
        Xn, Yn = np.meshgrid(x, y)
        Z = griddata(self.coordinates, values, (Xn, Yn), method="cubic")

        return Xn, Yn, Z
